# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import sciline
import scipp as sc

from ess import isissans as isis
from ess import sans
from ess.sans.types import (
    CorrectForGravity,
    Filename,
    Incident,
    IofQ,
    IofQxy,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    PixelMaskFilename,
    QBins,
    QxBins,
    QyBins,
    SampleRun,
    Transmission,
    UncertaintyBroadcastMode,
    WavelengthBins,
)


def make_params() -> dict:
    params = {
        **isis.default_parameters(),
        sans.types.DirectBeamFilename: isis.data.zoom_tutorial_direct_beam(),
        isis.CalibrationFilename: isis.data.zoom_tutorial_calibration(),
        Filename[sans.types.SampleRun]: isis.data.zoom_tutorial_sample_run(),
        Filename[sans.types.EmptyBeamRun]: isis.data.zoom_tutorial_empty_beam_run(),
        isis.SampleOffset: sc.vector([0.0, 0.0, 0.11], unit='m'),
        isis.DetectorBankOffset: sc.vector([0.0, 0.0, 0.5], unit='m'),
    }

    params[NeXusMonitorName[Incident]] = 'monitor3'
    params[NeXusMonitorName[Transmission]] = 'monitor5'

    params[WavelengthBins] = sc.geomspace(
        'wavelength', start=1.75, stop=16.5, num=141, unit='angstrom'
    )

    params[QBins] = sc.geomspace(
        dim='Q', start=0.004, stop=0.8, num=141, unit='1/angstrom'
    )

    params[NonBackgroundWavelengthRange] = sc.array(
        dims=['wavelength'], values=[0.7, 17.1], unit='angstrom'
    )
    params[CorrectForGravity] = True
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    params[sans.ReturnEvents] = False
    return params


def make_masks_table() -> sciline.ParamTable:
    masks = isis.data.zoom_tutorial_mask_filenames()
    return sciline.ParamTable(PixelMaskFilename, columns={}, index=masks)


def zoom_providers():
    return list(
        sans.providers
        + isis.providers
        + isis.mantidio.providers
        + (
            isis.data.transmission_from_background_run,
            isis.data.transmission_from_sample_run,
            isis.data.load_tutorial_direct_beam,
            isis.data.load_tutorial_run,
            sans.beam_center_finder.beam_center_from_center_of_mass,
        )
    )


def test_can_create_pipeline():
    pipeline = sciline.Pipeline(zoom_providers(), params=make_params())
    pipeline.set_param_table(make_masks_table())
    pipeline.get(IofQ[SampleRun])


def test_pipeline_can_compute_IofQ():
    pipeline = sciline.Pipeline(zoom_providers(), params=make_params())
    pipeline.set_param_table(make_masks_table())
    result = pipeline.compute(IofQ[SampleRun])
    assert result.dims == ('Q',)


def test_pipeline_can_compute_IofQxQy():
    pipeline = sciline.Pipeline(zoom_providers(), params=make_params())
    pipeline.set_param_table(make_masks_table())
    pipeline[QxBins] = sc.linspace(
        dim='Qx', start=-0.5, stop=0.5, num=101, unit='1/angstrom'
    )
    pipeline[QyBins] = sc.linspace(
        dim='Qy', start=-0.8, stop=0.8, num=101, unit='1/angstrom'
    )

    result = pipeline.compute(IofQxy[SampleRun])
    assert result.dims == ('Qy', 'Qx')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""Tools for handling statistical uncertainties."""

from typing import Dict, TypeVar, Union, overload

import scipp as sc

T = TypeVar("T", bound=Union[sc.Variable, sc.DataArray])


@overload
def broadcast_with_upper_bound_variances(
    data: sc.Variable, sizes: Dict[str, int]
) -> sc.Variable:
    pass


@overload
def broadcast_with_upper_bound_variances(
    data: sc.DataArray, sizes: Dict[str, int]
) -> sc.DataArray:
    pass


def broadcast_with_upper_bound_variances(
    data: Union[sc.Variable, sc.DataArray], sizes: Dict[str, int]
) -> Union[sc.Variable, sc.DataArray]:
    if _no_variance_broadcast(data, sizes):
        return data
    size = 1
    for dim, dim_size in sizes.items():
        if dim not in data.dims:
            size *= dim_size
    data = data.copy()
    data.variances *= size
    return data.broadcast(sizes={**sizes, **data.sizes}).copy()


def drop_variances_if_broadcast(
    data: Union[sc.Variable, sc.DataArray], sizes: Dict[str, int]
) -> Union[sc.Variable, sc.DataArray]:
    if _no_variance_broadcast(data, sizes):
        return data
    return sc.values(data)


def _no_variance_broadcast(
    data: Union[sc.Variable, sc.DataArray], sizes: Dict[str, int]
) -> bool:
    return (data.variances is None) or all(
        data.sizes.get(dim) == size for dim, size in sizes.items()
    )


def broadcast_to_events_with_upper_bound_variances(
    da: sc.DataArray, *, events: sc.DataArray
) -> sc.DataArray:
    """
    Upper-bound estimate for errors from normalization in event-mode.

    Count the number of events in each bin of the input data array. Then scale the
    variances by the number of events in each bin. An explicit broadcast is performed
    to bypass Scipp's safety check on broadcasting variances.

    Details will be published in an upcoming publication by Simon Heybrock et al.
    """
    if da.variances is None:
        return da
    if da.sizes != events.sizes:
        # This is a safety check, but we should never get here.
        raise ValueError(f"Sizes {da.sizes} do not match event sizes {events.sizes}")
    # Given how this function is used currently (in the context of normalization
    # with matching binning in numerator and denominator, not using scipp.lookup),
    # we can simply count the events in the existing binning.
    da.variances *= events.bins.size().values
    da.data = sc.bins_like(events, da.data)
    return da



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import sys
from pathlib import Path

import sciline
import scipp as sc
from scipp.scipy.interpolate import interp1d

from ess import loki, sans
from ess.sans.types import (
    DimsToKeep,
    PixelMaskFilename,
    QBins,
    WavelengthBands,
    WavelengthBins,
)

sys.path.insert(0, str(Path(__file__).resolve().parent))
from common import loki_providers, make_params  # noqa: E402


def _get_I0(qbins: sc.Variable) -> sc.Variable:
    Iq_theory = sc.io.load_hdf5(loki.data.loki_tutorial_poly_gauss_I0())
    f = interp1d(Iq_theory, 'Q')
    return f(sc.midpoints(qbins)).data[0]


def test_can_compute_direct_beam_for_all_pixels():
    n_wavelength_bands = 10
    params = make_params()
    params[WavelengthBands] = sc.linspace(
        'wavelength',
        params[WavelengthBins].min(),
        params[WavelengthBins].max(),
        n_wavelength_bands + 1,
    )
    providers = loki_providers()
    pipeline = sciline.Pipeline(providers, params=params)
    pipeline.set_param_series(PixelMaskFilename, [])
    I0 = _get_I0(qbins=params[QBins])

    results = sans.direct_beam(workflow=pipeline, I0=I0, niter=4)
    iofq_full = results[-1]['iofq_full']
    iofq_bands = results[-1]['iofq_bands']
    direct_beam_function = results[-1]['direct_beam']

    assert iofq_full.dims == ('Q',)
    assert iofq_bands.dims == ('band', 'Q')
    assert iofq_bands.sizes['band'] == n_wavelength_bands
    assert direct_beam_function.sizes['wavelength'] == n_wavelength_bands


def test_can_compute_direct_beam_with_overlapping_wavelength_bands():
    n_wavelength_bands = 10
    params = make_params()
    # Bands have double the width
    edges = sc.linspace(
        'band',
        params[WavelengthBins].min(),
        params[WavelengthBins].max(),
        n_wavelength_bands + 2,
    )
    params[WavelengthBands] = sc.concat(
        [edges[:-2], edges[2::]], dim='wavelength'
    ).transpose()

    providers = loki_providers()
    pipeline = sciline.Pipeline(providers, params=params)
    pipeline.set_param_series(PixelMaskFilename, [])
    I0 = _get_I0(qbins=params[QBins])

    results = sans.direct_beam(workflow=pipeline, I0=I0, niter=4)
    iofq_full = results[-1]['iofq_full']
    iofq_bands = results[-1]['iofq_bands']
    direct_beam_function = results[-1]['direct_beam']

    assert iofq_full.dims == ('Q',)
    assert iofq_bands.dims == ('band', 'Q')
    assert iofq_bands.sizes['band'] == n_wavelength_bands
    assert direct_beam_function.sizes['wavelength'] == n_wavelength_bands


def test_can_compute_direct_beam_per_layer():
    n_wavelength_bands = 10
    params = make_params()
    params[WavelengthBands] = sc.linspace(
        'wavelength',
        params[WavelengthBins].min(),
        params[WavelengthBins].max(),
        n_wavelength_bands + 1,
    )
    params[DimsToKeep] = ['layer']
    providers = loki_providers()
    pipeline = sciline.Pipeline(providers, params=params)
    pipeline.set_param_series(PixelMaskFilename, [])
    I0 = _get_I0(qbins=params[QBins])

    results = sans.direct_beam(workflow=pipeline, I0=I0, niter=4)
    iofq_full = results[-1]['iofq_full']
    iofq_bands = results[-1]['iofq_bands']
    direct_beam_function = results[-1]['direct_beam']

    assert iofq_full.dims == ('layer', 'Q')
    assert iofq_bands.dims == ('band', 'layer', 'Q')
    assert iofq_bands.sizes['band'] == n_wavelength_bands
    assert iofq_bands.sizes['layer'] == 4
    assert direct_beam_function.sizes['wavelength'] == n_wavelength_bands
    assert direct_beam_function.sizes['layer'] == 4


def test_can_compute_direct_beam_per_layer_and_straw():
    n_wavelength_bands = 10
    params = make_params()
    params[WavelengthBands] = sc.linspace(
        'wavelength',
        params[WavelengthBins].min(),
        params[WavelengthBins].max(),
        n_wavelength_bands + 1,
    )
    params[DimsToKeep] = ('layer', 'straw')
    providers = loki_providers()
    pipeline = sciline.Pipeline(providers, params=params)
    pipeline.set_param_series(PixelMaskFilename, [])
    I0 = _get_I0(qbins=params[QBins])

    results = sans.direct_beam(workflow=pipeline, I0=I0, niter=4)
    iofq_full = results[-1]['iofq_full']
    iofq_bands = results[-1]['iofq_bands']
    direct_beam_function = results[-1]['direct_beam']

    assert iofq_full.dims == ('layer', 'straw', 'Q')
    assert iofq_bands.dims == ('band', 'layer', 'straw', 'Q')
    assert iofq_bands.sizes['band'] == n_wavelength_bands
    assert iofq_bands.sizes['layer'] == 4
    assert iofq_bands.sizes['straw'] == 7
    assert direct_beam_function.sizes['wavelength'] == n_wavelength_bands
    assert direct_beam_function.sizes['layer'] == 4
    assert direct_beam_function.sizes['straw'] == 7



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import sciline

from ess.sans import providers as sans_providers

from .data import load_tutorial_direct_beam, load_tutorial_run
from .general import default_parameters
from .io import read_xml_detector_masking
from .mantidio import providers as mantid_providers


def set_mantid_log_level(level: int = 3):
    try:
        from mantid import ConfigService

        cfg = ConfigService.Instance()
        cfg.setLogLevel(level)  # Silence verbose load via Mantid
    except ImportError:
        pass


def ZoomWorkflow() -> sciline.Pipeline:
    """Create Zoom workflow with default parameters."""
    from . import providers as isis_providers

    set_mantid_log_level()

    params = default_parameters()
    zoom_providers = sans_providers + isis_providers + mantid_providers
    workflow = sciline.Pipeline(providers=zoom_providers, params=params)
    workflow.insert(read_xml_detector_masking)
    return workflow


def ZoomTutorialWorkflow() -> sciline.Pipeline:
    """
    Create Zoom tutorial workflow.

    Equivalent to :func:`ZoomWorkflow`, but with loaders for tutorial data instead
    of Mantid-based loaders.
    """
    workflow = ZoomWorkflow()
    workflow.insert(load_tutorial_run)
    workflow.insert(load_tutorial_direct_beam)
    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Optional

import scipp as sc
from scipp.core import concepts

from .types import (
    CalibratedMaskedData,
    CleanDirectBeam,
    CleanMonitor,
    CleanWavelength,
    Denominator,
    DetectorPixelShape,
    EmptyBeamRun,
    FinalSummedQ,
    FinalSummedQxy,
    Incident,
    IofQ,
    IofQxy,
    LabFrameTransform,
    NormWavelengthTerm,
    Numerator,
    ProcessedWavelengthBands,
    ReturnEvents,
    ScatteringRunType,
    SolidAngle,
    Transmission,
    TransmissionFraction,
    TransmissionRun,
    UncertaintyBroadcastMode,
    WavelengthBands,
    WavelengthBins,
)
from .uncertainty import (
    broadcast_to_events_with_upper_bound_variances,
    broadcast_with_upper_bound_variances,
    drop_variances_if_broadcast,
)


def solid_angle(
    data: CalibratedMaskedData[ScatteringRunType],
    pixel_shape: DetectorPixelShape[ScatteringRunType],
    transform: LabFrameTransform[ScatteringRunType],
) -> SolidAngle[ScatteringRunType]:
    """
    Solid angle for cylindrical pixels.

    Note that the approximation is valid when the distance from sample
    to pixel is much larger than the length or the radius of the pixels.

    Parameters
    ----------
    data:
        The DataArray that contains the positions of the detector pixels
        and the position of the sample in the coords.

    pixel_shape:
        Contains the description of the detector pixel shape.

    transform:
        Transformation from the local coordinate system of the detector
        to the coordinate system of the sample.

    Returns
    -------
    :
        The solid angle of the detector pixels, as viewed from the sample position.
    """
    face_1_center, face_1_edge, face_2_center = (
        pixel_shape['vertices']['vertex', i] for i in range(3)
    )
    cylinder_axis = transform * face_2_center - transform * face_1_center
    radius = sc.norm(face_1_center - face_1_edge)
    length = sc.norm(cylinder_axis)

    omega = _approximate_solid_angle_for_cylinder_shaped_pixel_of_detector(
        pixel_position=data.coords['position'] - data.coords['sample_position'],
        cylinder_axis=cylinder_axis,
        radius=radius,
        length=length,
    )
    return SolidAngle[ScatteringRunType](
        concepts.rewrap_reduced_data(
            prototype=data, data=omega, dim=set(data.dims) - set(omega.dims)
        )
    )


def _approximate_solid_angle_for_cylinder_shaped_pixel_of_detector(
    pixel_position: sc.Variable,
    cylinder_axis: sc.Variable,
    radius: sc.Variable,
    length: sc.Variable,
):
    r"""Computes solid angle of a detector pixel under the assumption that the
    distance to the detector pixel from the sample is large compared to the radius
    and to the length of the piece of the detector cylinder that makes up the pixel.

    The cylinder is approximated by a rectangle element contained in the cylinder.
    The normal of the rectangle is selected to
    1. be orthogonal to the cylinder axis and
    2. "maximally parallel" to the pixel position vector.
    This occurs when the normal :math:`n`, the cylinder axis :math:`c` and the
    pixel position :math:`r` fall in a shared plane such that
    .. math::
        r = (n\cdot r)n + (c \cdot r)c.

    From the above relationship we get the scalar product of the position
    vector and the normal:

    .. math::
        |\hat{r}\cdot n| = \sqrt{\frac{r \cdot r - (c \cdot r^2)}{||r||^2}}

    The solid angle contribution from the detector element is approximated as
    .. math::
        \Delta\Omega =  A \frac{|\hat{r}\cdot n|}{||r||^2}
    where :math:`A = 2 R L` is the area of the rectangular element
    and :math:`\hat{r}` is the normalized pixel position vector.
    """
    norm_pp = sc.norm(pixel_position)
    norm_ca = sc.norm(cylinder_axis)
    cosalpha = sc.sqrt(
        1 - (sc.dot(pixel_position, cylinder_axis) / (norm_pp * norm_ca)) ** 2
    )
    return (2 * radius * length) * cosalpha / norm_pp**2


def transmission_fraction(
    sample_incident_monitor: CleanMonitor[TransmissionRun[ScatteringRunType], Incident],
    sample_transmission_monitor: CleanMonitor[
        TransmissionRun[ScatteringRunType], Transmission
    ],
    direct_incident_monitor: CleanMonitor[EmptyBeamRun, Incident],
    direct_transmission_monitor: CleanMonitor[EmptyBeamRun, Transmission],
) -> TransmissionFraction[ScatteringRunType]:
    """
    Approximation based on equations in
    `CalculateTransmission <https://docs.mantidproject.org/v4.0.0/algorithms/CalculateTransmission-v1.html>`_
    documentation:
    ``(sample_transmission_monitor / direct_transmission_monitor) * (direct_incident_monitor / sample_incident_monitor)``

    This is equivalent to ``mantid.CalculateTransmission`` without fitting.
    Inputs should be wavelength-dependent.

    Parameters
    ----------
    sample_incident_monitor:
        The incident monitor data for the sample (transmission) run.
    sample_transmission_monitor:
        The transmission monitor data for the sample (transmission) run.
    direct_incident_monitor:
        The incident monitor data for the direct beam run.
    direct_transmission_monitor:
        The transmission monitor data for the direct beam run.

    Returns
    -------
    :
        The transmission fraction computed from the monitor counts.
    """  # noqa: E501
    frac = (sample_transmission_monitor / direct_transmission_monitor) * (
        direct_incident_monitor / sample_incident_monitor
    )
    return TransmissionFraction[ScatteringRunType](frac)


_broadcasters = {
    UncertaintyBroadcastMode.drop: drop_variances_if_broadcast,
    UncertaintyBroadcastMode.upper_bound: broadcast_with_upper_bound_variances,
    UncertaintyBroadcastMode.fail: lambda x, sizes: x,
}


def iofq_norm_wavelength_term(
    incident_monitor: CleanMonitor[ScatteringRunType, Incident],
    transmission_fraction: TransmissionFraction[ScatteringRunType],
    direct_beam: Optional[CleanDirectBeam],
    uncertainties: UncertaintyBroadcastMode,
) -> NormWavelengthTerm[ScatteringRunType]:
    """
    Compute the wavelength-dependent contribution to the denominator term for the I(Q)
    normalization.
    Keeping this as a separate function allows us to compute it once during the
    iterations for finding the beam center, while the solid angle is recomputed
    for each iteration.

    This is basically:
    ``incident_monitor * transmission_fraction * direct_beam``
    If the direct beam is not supplied, it is assumed to be 1.

    Because the multiplication between the ``incident_monitor * transmission_fraction``
    (pixel-independent) and the direct beam (potentially pixel-dependent) consists of a
    broadcast operation which would introduce correlations, variances of the direct
    beam are dropped or replaced by an upper-bound estimation, depending on the
    configured mode.

    Parameters
    ----------
    incident_monitor:
        The incident monitor data (depends on wavelength).
    transmission_fraction:
        The transmission fraction (depends on wavelength).
    direct_beam:
        The direct beam function (depends on wavelength).
    uncertainties:
        The mode for broadcasting uncertainties. See
        :py:class:`UncertaintyBroadcastMode` for details.

    Returns
    -------
    :
        Wavelength-dependent term
        (incident_monitor * transmission_fraction * direct_beam) to be used for
        the denominator of the SANS I(Q) normalization.
        Used by :py:func:`iofq_denominator`.
    """
    out = incident_monitor * transmission_fraction
    if direct_beam is not None:
        # Make wavelength the inner dim
        dims = list(direct_beam.dims)
        dims.remove('wavelength')
        dims.append('wavelength')
        direct_beam = direct_beam.transpose(dims)
        broadcast = _broadcasters[uncertainties]
        out = direct_beam * broadcast(out, sizes=direct_beam.sizes)
    # Convert wavelength coordinate to midpoints for future histogramming
    out.coords['wavelength'] = sc.midpoints(out.coords['wavelength'])
    return NormWavelengthTerm[ScatteringRunType](out)


def iofq_denominator(
    wavelength_term: NormWavelengthTerm[ScatteringRunType],
    solid_angle: SolidAngle[ScatteringRunType],
    uncertainties: UncertaintyBroadcastMode,
) -> CleanWavelength[ScatteringRunType, Denominator]:
    """
    Compute the denominator term for the I(Q) normalization.

    In a SANS experiment, the scattering cross section :math:`I(Q)` is defined as
    (`Heenan et al. 1997 <https://doi.org/10.1107/S0021889897002173>`_):

    .. math::

       I(Q) = \\frac{\\partial\\Sigma{Q}}{\\partial\\Omega} = \\frac{A_{H} \\Sigma_{R,\\lambda\\subset Q} C(R, \\lambda)}{A_{M} t \\Sigma_{R,\\lambda\\subset Q}M(\\lambda)T(\\lambda)D(\\lambda)\\Omega(R)}

    where :math:`A_{H}` is the area of a mask (which avoids saturating the detector)
    placed between the monitor of area :math:`A_{M}` and the main detector.
    :math:`\\Omega` is the detector solid angle, and :math:`C` is the count rate on the
    main detector, which depends on the position :math:`R` and the wavelength.
    :math:`t` is the sample thickness, :math:`M` represents the incident monitor count
    rate for the sample run, and :math:`T` is known as the transmission fraction.

    Note that the incident monitor used to compute the transmission fraction is not
    necessarily the same as :math:`M`, as the transmission fraction is usually computed
    from a separate 'transmission' run (in the 'sample' run, the transmission monitor is
    commonly moved out of the beam path, to avoid polluting the sample detector signal).

    Finally, :math:`D` is the 'direct beam function', and is defined as

    .. math::

       D(\\lambda) = \\frac{\\eta(\\lambda)}{\\eta_{M}(\\lambda)} \\frac{A_{H}}{A_{M}}

    where :math:`\\eta` and :math:`\\eta_{M}` are the detector and monitor
    efficiencies, respectively.

    Hence, in order to normalize the main detector counts :math:`C`, we need compute the
    transmission fraction :math:`T(\\lambda)`, the direct beam function
    :math:`D(\\lambda)` and the solid angle :math:`\\Omega(R)`.

    The denominator is then simply:
    :math:`M_{\\lambda} T_{\\lambda} D_{\\lambda} \\Omega_{R}`,
    which is equivalent to ``wavelength_term * solid_angle``.
    The ``wavelength_term`` includes all but the ``solid_angle`` and is computed by
    :py:func:`iofq_norm_wavelength_term_sample` or
    :py:func:`iofq_norm_wavelength_term_background`.

    Because the multiplication between the wavelength dependent terms
    and the pixel dependent term (solid angle) consists of a broadcast operation which
    would introduce correlations, variances are dropped or replaced by an upper-bound
    estimation, depending on the configured mode.

    Parameters
    ----------
    wavelength_term:
        The term that depends on wavelength, computed by
        :py:func:`iofq_norm_wavelength_term`.
    solid_angle:
        The solid angle of the detector pixels, as viewed from the sample position.
    uncertainties:
        The mode for broadcasting uncertainties. See
        :py:class:`UncertaintyBroadcastMode` for details.

    Returns
    -------
    :
        The denominator for the SANS I(Q) normalization.
    """  # noqa: E501
    broadcast = _broadcasters[uncertainties]
    denominator = solid_angle * broadcast(wavelength_term, sizes=solid_angle.sizes)
    return CleanWavelength[ScatteringRunType, Denominator](denominator)


def process_wavelength_bands(
    wavelength_bands: Optional[WavelengthBands],
    wavelength_bins: WavelengthBins,
) -> ProcessedWavelengthBands:
    """
    Perform some checks and potential reshaping on the wavelength bands.

    The wavelength bands must be either one- or two-dimensional.
    If the wavelength bands are defined as a one-dimensional array, convert them to a
    two-dimensional array with start and end wavelengths.

    The final bands must have a size of 2 in the wavelength dimension, defining a start
    and an end wavelength.
    """
    if wavelength_bands is None:
        wavelength_bands = sc.concat(
            [wavelength_bins.min(), wavelength_bins.max()], dim='wavelength'
        )
    if wavelength_bands.ndim == 1:
        wavelength_bands = sc.concat(
            [wavelength_bands[:-1], wavelength_bands[1:]], dim='x'
        ).rename(x='wavelength', wavelength='band')
    if wavelength_bands.ndim != 2:
        raise ValueError(
            'Wavelength_bands must be one- or two-dimensional, '
            f'got {wavelength_bands.ndim}.'
        )
    if wavelength_bands.sizes['wavelength'] != 2:
        raise ValueError(
            'Wavelength_bands must have a size of 2 in the wavelength dimension, '
            'defining a start and an end wavelength, '
            f'got {wavelength_bands.sizes["wavelength"]}.'
        )
    return wavelength_bands


def _normalize(
    numerator: sc.DataArray,
    denominator: sc.DataArray,
    return_events: ReturnEvents,
    uncertainties: UncertaintyBroadcastMode,
    wavelength_bands: ProcessedWavelengthBands,
) -> sc.DataArray:
    """
    Perform normalization of counts as a function of Q.
    If the numerator contains events, we use the sc.lookup function to perform the
    division.

    Parameters
    ----------
    numerator:
        The data whose counts will be divided by the denominator. This can either be
        event or dense (histogrammed) data.
    denominator:
        The divisor for the normalization operation. This cannot be event data, it must
        contain histogrammed data.
    return_events:
        Whether to return the result as event data or histogrammed data.
    wavelength_bands:
        Defines bands in wavelength that can be used to separate different wavelength
        ranges that contribute to different regions in Q space. Note that this needs to
        be defined, so if all wavelengths should be used, this should simply be a start
        and end edges that encompass the entire wavelength range.

    Returns
    -------
    :
        The input data normalized by the supplied denominator.
    """
    wav = 'wavelength'
    wavelength_bounds = sc.sort(wavelength_bands.flatten(to=wav), wav)
    if numerator.bins is not None:
        # If in event mode the desired wavelength binning has not been applied, we need
        # it for splitting by bands, or restricting the range in case of a single band.
        numerator = numerator.bin(wavelength=wavelength_bounds)

    def _reduce(da: sc.DataArray) -> sc.DataArray:
        if da.sizes[wav] == 1:  # Can avoid costly event-data da.bins.concat
            return da.squeeze(wav)
        return da.sum(wav) if da.bins is None else da.bins.concat(wav)

    num_parts = []
    denom_parts = []
    for wav_range in sc.collapse(wavelength_bands, keep=wav).values():
        num_parts.append(_reduce(numerator[wav, wav_range[0] : wav_range[1]]))
        denom_parts.append(_reduce(denominator[wav, wav_range[0] : wav_range[1]]))
    band_dim = (set(wavelength_bands.dims) - {'wavelength'}).pop()
    if len(num_parts) == 1:
        numerator = num_parts[0]
        denominator = denom_parts[0]
    else:
        numerator = sc.concat(num_parts, band_dim)
        denominator = sc.concat(denom_parts, band_dim)
    numerator.coords[wav] = wavelength_bands.squeeze()
    denominator.coords[wav] = wavelength_bands.squeeze()

    if return_events and numerator.bins is not None:
        # Naive event-mode normalization is not correct if norm-term has variances.
        # See https://doi.org/10.3233/JNR-220049 for context.
        if denominator.variances is not None:
            if uncertainties == UncertaintyBroadcastMode.drop:
                denominator = sc.values(denominator)
            else:
                denominator = broadcast_to_events_with_upper_bound_variances(
                    denominator, events=numerator
                )
    elif numerator.bins is not None:
        numerator = numerator.hist()
    numerator /= denominator.drop_coords(
        [name for name in denominator.coords if name not in denominator.dims]
    )
    return numerator


def normalize_q(
    numerator: FinalSummedQ[ScatteringRunType, Numerator],
    denominator: FinalSummedQ[ScatteringRunType, Denominator],
    return_events: ReturnEvents,
    uncertainties: UncertaintyBroadcastMode,
    wavelength_bands: ProcessedWavelengthBands,
) -> IofQ[ScatteringRunType]:
    return IofQ[ScatteringRunType](
        _normalize(
            numerator=numerator,
            denominator=denominator,
            return_events=return_events,
            uncertainties=uncertainties,
            wavelength_bands=wavelength_bands,
        )
    )


def normalize_qxy(
    numerator: FinalSummedQxy[ScatteringRunType, Numerator],
    denominator: FinalSummedQxy[ScatteringRunType, Denominator],
    return_events: ReturnEvents,
    uncertainties: UncertaintyBroadcastMode,
    wavelength_bands: ProcessedWavelengthBands,
) -> IofQxy[ScatteringRunType]:
    return IofQxy[ScatteringRunType](
        _normalize(
            numerator=numerator,
            denominator=denominator,
            return_events=return_events,
            uncertainties=uncertainties,
            wavelength_bands=wavelength_bands,
        )
    )


normalize_q.__doc__ = _normalize.__doc__
normalize_qxy.__doc__ = _normalize.__doc__


providers = (
    transmission_fraction,
    iofq_norm_wavelength_term,
    iofq_denominator,
    normalize_q,
    normalize_qxy,
    process_wavelength_bands,
    solid_angle,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""
Plotting functions for ISIS data.
"""
import warnings
from typing import Any

import scipp as sc


def plot_flat_detector_xy(
    da: sc.DataArray, pixels_per_tube: int = 512, **kwargs: Any
) -> Any:
    """
    Plot a 2-D view of data from a single flat detector bank.

    This is an alternative to the `scn.instrument_view` function, avoiding the 3-D
    rendering of the instrument. The exact X and Y coordinates of the pixels are
    used for the 2-D plot.

    There are currently no checks that the data array is actually from a single
    detector bank, or that it is flat. The caller is responsible for ensuring this.

    Parameters
    ----------
    da:
        The data array to plot. Must have a 'position' coord and a single dimension.
    pixels_per_tube:
        The number of pixels per tube. Defaults to 512.
    kwargs:
        Additional arguments passed to `sc.plot`.
    """
    if da.bins is not None:
        da = da.hist()
    da.coords['x'] = da.coords['position'].fields.x.copy()
    da.coords['y'] = da.coords['position'].fields.y.copy()
    folded = da.fold(da.dim, sizes={'y': -1, 'x': pixels_per_tube})
    y = folded.coords['y']
    if sc.all(y.min('x') == y.max('x')):
        folded.coords['y'] = y.min('x')
    else:
        raise ValueError(
            'Cannot plot 2-D instrument view of data array with non-constant '
            'y coordinate along tubes. Use scippneutron.instrument_view instead.'
        )
    plot_kwargs = dict(aspect='equal')
    plot_kwargs.update(kwargs)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        out = folded.plot(**plot_kwargs)
    return out



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc

from ess.isissans.data import sans2d_solid_angle_reference
from ess.sans import normalization

# See https://github.com/mantidproject/mantid/blob/main/instrument/SANS2D_Definition_Tubes.xml  # noqa: E501
_SANS2D_PIXEL_RADIUS = 0.00405 * sc.Unit('m')
_SANS2D_PIXEL_LENGTH = 0.002033984375 * sc.Unit('m')


def _sans2d_geometry():
    R = _SANS2D_PIXEL_RADIUS.value
    L = _SANS2D_PIXEL_LENGTH.value
    pixel_shape = {
        'vertices': sc.vectors(
            dims=['vertex'],
            values=[
                [0, 0, 0],
                [R, 0, 0],
                [0, L, 0],
            ],
            unit='m',
        )
    }
    # Rotate +y to -x
    transform = sc.spatial.rotation(value=[0, 0, 1 / 2**0.5, 1 / 2**0.5])
    return dict(pixel_shape=pixel_shape, transform=transform)


def _mantid_sans2d_solid_angle_data():
    simpleapi = pytest.importorskip("mantid.simpleapi")
    scippneutron = pytest.importorskip("scippneutron")

    ws = simpleapi.Load('SANS2D00063091.nxs')
    radius = _SANS2D_PIXEL_RADIUS
    length = _SANS2D_PIXEL_LENGTH

    simpleapi.SetInstrumentParameter(
        ws,
        ParameterName='x-pixel-size',
        ParameterType='Number',
        Value=str((2 * radius).to(unit='mm').value),
    )
    simpleapi.SetInstrumentParameter(
        ws,
        ParameterName='y-pixel-size',
        ParameterType='Number',
        Value=str(length.to(unit='mm').value),
    )
    outWs = simpleapi.SolidAngle(ws, method='HorizontalTube')
    da = scippneutron.from_mantid(outWs)['data']['spectrum', :120000:100]
    # Create new reference file:
    # sc.io.hdf5.save_hdf5(da, 'SANS2D00063091.SolidAngle_from_mantid.h5')
    # Note, also update the registry version, don't overwrite old reference files.
    # Overwriting reference files breaks old versions of the repository.
    return da


@pytest.mark.filterwarnings("ignore:.*")
def test_solid_angle_compare_to_mantid():
    da = _mantid_sans2d_solid_angle_data()
    solid_angle = normalization.solid_angle(
        da,
        **_sans2d_geometry(),
    ).data
    assert sc.allclose(
        da.data['tof', 0], solid_angle, atol=0.0 * sc.Unit('dimensionless')
    )


def test_solid_angle_compare_to_reference_file():
    da = sc.io.load_hdf5(filename=sans2d_solid_angle_reference())
    solid_angle = normalization.solid_angle(
        da,
        **_sans2d_geometry(),
    ).data
    assert sc.allclose(
        da.data['tof', 0], solid_angle, atol=0.0 * sc.Unit('dimensionless')
    )


def test_transmission_fraction():
    N = 100
    wavelength = sc.linspace(
        dim='wavelength', start=2.0, stop=16.0, num=N + 1, unit='angstrom'
    )
    sample_incident_monitor = sc.DataArray(
        data=sc.array(
            dims=['wavelength'], values=100.0 * np.random.random(N), unit='counts'
        ),
        coords={'wavelength': wavelength},
    )
    sample_transmission_monitor = sc.DataArray(
        data=sc.array(
            dims=['wavelength'], values=50.0 * np.random.random(N), unit='counts'
        ),
        coords={'wavelength': wavelength},
    )

    direct_incident_monitor = sc.DataArray(
        data=sc.array(
            dims=['wavelength'], values=100.0 * np.random.random(N), unit='counts'
        ),
        coords={'wavelength': wavelength},
    )
    direct_transmission_monitor = sc.DataArray(
        data=sc.array(
            dims=['wavelength'], values=80.0 * np.random.random(N), unit='counts'
        ),
        coords={'wavelength': wavelength},
    )

    trans_frac = normalization.transmission_fraction(
        sample_incident_monitor=sample_incident_monitor,
        sample_transmission_monitor=sample_transmission_monitor,
        direct_incident_monitor=direct_incident_monitor,
        direct_transmission_monitor=direct_transmission_monitor,
    )

    # If counts on data transmission monitor have increased, it means less neutrons
    # have been absorbed and transmission fraction should increase.
    # - data run: incident: 100 -> transmission: 75
    # - direct run: incident: 100 -> transmission: 80
    assert sc.allclose(
        (trans_frac * sc.scalar(1.5)).data,
        normalization.transmission_fraction(
            sample_incident_monitor=sample_incident_monitor,
            sample_transmission_monitor=sample_transmission_monitor * sc.scalar(1.5),
            direct_incident_monitor=direct_incident_monitor,
            direct_transmission_monitor=direct_transmission_monitor,
        ).data,
    )

    # If counts on direct transmission monitor are higher, it means that many more
    # neutrons are absorbed when the sample is in the path of the beam, and therefore
    # the transmission fraction should decrease.
    # - data run: incident: 100 -> transmission: 50
    # - direct run: incident: 100 -> transmission: 90
    assert sc.allclose(
        (trans_frac / sc.scalar(9 / 8)).data,
        normalization.transmission_fraction(
            sample_incident_monitor=sample_incident_monitor,
            sample_transmission_monitor=sample_transmission_monitor,
            direct_incident_monitor=direct_incident_monitor,
            direct_transmission_monitor=direct_transmission_monitor * sc.scalar(9 / 8),
        ).data,
    )

    # If counts on direct incident monitor are higher, but counts on direct transmission
    # monitor are the same, it means that the relative difference between incident and
    # transmission has increased for the direct run, but not for the data run.
    # This would be the case where neutron beam flux was higher during the direct run.
    # This implies that that the transmission fraction is higher than in our vanilla
    # run.
    # - data run: incident: 100 -> transmission: 50
    # - direct run: incident: 110 -> transmission: 80
    assert sc.allclose(
        (trans_frac * sc.scalar(1.1)).data,
        normalization.transmission_fraction(
            sample_incident_monitor=sample_incident_monitor,
            sample_transmission_monitor=sample_transmission_monitor,
            direct_incident_monitor=direct_incident_monitor * sc.scalar(1.1),
            direct_transmission_monitor=direct_transmission_monitor,
        ).data,
    )

    # If counts on data incident monitor are higher, but counts on data transmission
    # monitor are the same, it means that more neutrons were absorbed in this run,
    # and then transmission fraction decreases.
    # - data run: incident: 110 -> transmission: 50
    # - direct run: incident: 100 -> transmission: 80
    assert sc.allclose(
        (trans_frac / sc.scalar(1.1)).data,
        normalization.transmission_fraction(
            sample_incident_monitor=sample_incident_monitor * sc.scalar(1.1),
            sample_transmission_monitor=sample_transmission_monitor,
            direct_incident_monitor=direct_incident_monitor,
            direct_transmission_monitor=direct_transmission_monitor,
        ).data,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""
File loading functions for ISIS data using Mantid.
"""
from typing import NewType, NoReturn, Optional

import sciline
import scipp as sc
import scippneutron as scn
from scipp.constants import g

from ess.sans.types import DirectBeam, DirectBeamFilename, Filename, RunType, SampleRun

from .data import LoadedFileContents
from .io import CalibrationFilename

try:
    import mantid.api as _mantid_api
    import mantid.simpleapi as _mantid_simpleapi
    from mantid.api import MatrixWorkspace
except ModuleNotFoundError:
    # Catch runtime usages of Mantid
    class _MantidFallback:
        def __getattr__(self, name: str) -> NoReturn:
            raise ImportError(
                'Mantid is required to use `sans.isis.mantidio` but is not installed'
            ) from None

    _mantid_api = _MantidFallback()
    _mantid_simpleapi = _MantidFallback
    # Needed for type annotations
    MatrixWorkspace = object

Period = NewType('Period', int)
"""Period number of the events."""

CalibrationWorkspace = NewType('CalibrationWorkspace', MatrixWorkspace)


class DataWorkspace(sciline.Scope[RunType, MatrixWorkspace], MatrixWorkspace):
    """Workspace containing data"""


def _make_detector_info(ws: MatrixWorkspace) -> sc.DataGroup:
    det_info = scn.mantid.make_detector_info(ws, 'spectrum')
    return sc.DataGroup(det_info.coords)


def _get_detector_ids(ws: DataWorkspace[SampleRun]) -> sc.Variable:
    det_info = _make_detector_info(ws)
    dim = 'spectrum'
    da = sc.DataArray(det_info['detector'], coords={dim: det_info[dim]})
    da = sc.sort(da, dim)  # sort by spectrum index
    if not sc.identical(
        da.coords[dim],
        sc.arange('detector', da.sizes['detector'], dtype='int32', unit=None),
    ):
        raise ValueError("Spectrum-detector mapping is not 1:1, this is not supported.")
    return da.data.rename_dims(detector='spectrum')


def load_calibration(filename: CalibrationFilename) -> CalibrationWorkspace:
    ws = _mantid_simpleapi.Load(Filename=str(filename), StoreInADS=False)
    return CalibrationWorkspace(ws)


def load_direct_beam(filename: DirectBeamFilename) -> DirectBeam:
    dg = scn.load_with_mantid(
        filename=filename,
        mantid_alg="LoadRKH",
        mantid_args={"FirstColumnValue": "Wavelength"},
    )
    da = dg['data']
    del da.coords['spectrum']
    return DirectBeam(da)


def from_data_workspace(
    ws: DataWorkspace[RunType], calibration: Optional[CalibrationWorkspace]
) -> LoadedFileContents[RunType]:
    if calibration is not None:
        _mantid_simpleapi.CopyInstrumentParameters(
            InputWorkspace=calibration, OutputWorkspace=ws, StoreInADS=False
        )
    up = ws.getInstrument().getReferenceFrame().vecPointingUp()
    dg = scn.from_mantid(ws)
    det_ids = _get_detector_ids(ws)
    # In some instruments (e.g. SANS2D), some pixels are used for other purposes (e.g.
    # live acquisition). They have no detector ids, so we exclude them from the data.
    for dim, shape in det_ids.sizes.items():
        dg['data'] = dg['data'][dim, :shape]
    dg['data'] = dg['data'].squeeze()
    if (dg['data'].bins is not None) and ('tof' in dg['data'].coords):
        del dg['data'].coords['tof']
    dg['data'].coords['detector_id'] = det_ids
    dg['data'].coords['gravity'] = sc.vector(value=-up) * g
    return LoadedFileContents[RunType](dg)


def load_run(
    filename: Filename[RunType], period: Optional[Period]
) -> DataWorkspace[RunType]:
    loaded = _mantid_simpleapi.Load(
        Filename=str(filename), LoadMonitors=True, StoreInADS=False
    )
    if isinstance(loaded, _mantid_api.Workspace):
        # A single workspace
        data_ws = loaded
    else:
        # Separate data and monitor workspaces
        data_ws = loaded.OutputWorkspace
        if isinstance(data_ws, _mantid_api.WorkspaceGroup):
            if period is None:
                raise ValueError(
                    f'Needs {Period} to be set to know what '
                    'section of the event data to load'
                )
            data_ws = data_ws.getItem(period)
            data_ws.setMonitorWorkspace(loaded.MonitorWorkspace.getItem(period))
        else:
            data_ws.setMonitorWorkspace(loaded.MonitorWorkspace)
    return DataWorkspace[RunType](data_ws)


providers = (
    from_data_workspace,
    load_calibration,
    load_direct_beam,
    load_run,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
Loading and merging of LoKI data.
"""
import scipp as sc
from ess.reduce import nexus

from ess.sans.types import (
    Filename,
    LoadedNeXusDetector,
    LoadedNeXusMonitor,
    MonitorType,
    NeXusDetectorName,
    NeXusMonitorName,
    RawSample,
    RawSource,
    RunType,
)


def load_nexus_sample(file_path: Filename[RunType]) -> RawSample[RunType]:
    return RawSample[RunType](nexus.load_sample(file_path))


def dummy_load_sample(file_path: Filename[RunType]) -> RawSample[RunType]:
    return RawSample[RunType](
        sc.DataGroup({'position': sc.vector(value=[0, 0, 0], unit='m')})
    )


def load_nexus_source(file_path: Filename[RunType]) -> RawSource[RunType]:
    return RawSource[RunType](nexus.load_source(file_path))


def load_nexus_detector(
    file_path: Filename[RunType], detector_name: NeXusDetectorName
) -> LoadedNeXusDetector[RunType]:
    return LoadedNeXusDetector[RunType](
        nexus.load_detector(file_path=file_path, detector_name=detector_name)
    )


def load_nexus_monitor(
    file_path: Filename[RunType],
    monitor_name: NeXusMonitorName[MonitorType],
) -> LoadedNeXusMonitor[RunType, MonitorType]:
    return LoadedNeXusMonitor[RunType, MonitorType](
        nexus.load_monitor(file_path=file_path, monitor_name=monitor_name)
    )


providers = (
    load_nexus_detector,
    load_nexus_monitor,
    load_nexus_sample,
    load_nexus_source,
)
"""Providers for loading single files."""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import scipp as sc
import scippnexus as snx
from scippnexus.application_definitions import nxcansas

from .types import BackgroundSubtractedIofQ, OutFilename, RunNumber, RunTitle


def save_background_subtracted_iofq(
    *,
    iofq: BackgroundSubtractedIofQ,
    out_filename: OutFilename,
    run_number: RunNumber,
    run_title: RunTitle,
) -> None:
    """Save background-subtracted I(Q) histogram as an NXcanSAS file."""
    if iofq.bins is None:
        da = iofq.copy(deep=False)
    else:
        da = iofq.hist()
    if da.coords.is_edges('Q'):
        da.coords['Q'] = sc.midpoints(da.coords['Q'])
    with snx.File(out_filename, 'w') as f:
        f['sasentry'] = nxcansas.SASentry(title=run_title, run=run_number)
        f['sasentry']['sasdata'] = nxcansas.SASdata(da, Q_variances='resolutions')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
Masking functions for the loki workflow.
"""
import numpy as np
import sciline
import scipp as sc

from .types import (
    MaskedData,
    MaskedDetectorIDs,
    PixelMaskFilename,
    ScatteringRunType,
    TofData,
)


def apply_pixel_masks(
    data: TofData[ScatteringRunType],
    masked_ids: sciline.Series[PixelMaskFilename, MaskedDetectorIDs],
) -> MaskedData[ScatteringRunType]:
    """Apply pixel-specific masks to raw data.
    The masks are based on detector IDs stored in XML files.

    Parameters
    ----------
    data:
        Raw data with configured component positions.
    masks:
        A series of masks.
    """
    if masked_ids is not None:
        data = data.copy(deep=False)
        ids = data.coords[
            'detector_number' if 'detector_number' in data.coords else 'detector_id'
        ]
        for name, masked in masked_ids.items():
            mask = sc.zeros(sizes=ids.sizes, dtype='bool')
            mask.values[np.isin(ids.values, masked.values)] = True
            data.masks[name] = mask
    return MaskedData[ScatteringRunType](data)


providers = (apply_pixel_masks,)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from typing import NewType, Optional

import sciline
import scipp as sc

from ess.sans import providers as sans_providers
from ess.sans.types import MaskedData, SampleRun, ScatteringRunType, TofData

from .data import load_tutorial_direct_beam, load_tutorial_run
from .general import default_parameters
from .mantidio import providers as mantid_providers

DetectorEdgeMask = NewType('DetectorEdgeMask', sc.Variable)
"""Detector edge mask"""

LowCountThreshold = NewType('LowCountThreshold', sc.Variable)
"""Threshold below which detector pixels should be masked
(low-counts on the edges of the detector panel, and the beam stop)"""

SampleHolderMask = NewType('SampleHolderMask', sc.Variable)
"""Sample holder mask"""


def detector_edge_mask(
    sample: TofData[SampleRun],
) -> DetectorEdgeMask:
    mask_edges = (
        sc.abs(sample.coords['position'].fields.x) > sc.scalar(0.48, unit='m')
    ) | (sc.abs(sample.coords['position'].fields.y) > sc.scalar(0.45, unit='m'))
    return DetectorEdgeMask(mask_edges)


def sample_holder_mask(
    sample: TofData[SampleRun],
    low_counts_threshold: LowCountThreshold,
) -> SampleHolderMask:
    summed = sample.hist()
    holder_mask = (
        (summed.data < low_counts_threshold)
        & (sample.coords['position'].fields.x > sc.scalar(0, unit='m'))
        & (sample.coords['position'].fields.x < sc.scalar(0.42, unit='m'))
        & (sample.coords['position'].fields.y < sc.scalar(0.05, unit='m'))
        & (sample.coords['position'].fields.y > sc.scalar(-0.15, unit='m'))
    )
    return SampleHolderMask(holder_mask)


def mask_detectors(
    da: TofData[ScatteringRunType],
    edge_mask: Optional[DetectorEdgeMask],
    holder_mask: Optional[SampleHolderMask],
) -> MaskedData[ScatteringRunType]:
    """Apply pixel-specific masks to raw data.

    Parameters
    ----------
    da:
        Raw data.
    edge_mask:
        Mask for detector edges.
    holder_mask:
        Mask for sample holder.
    """
    da = da.copy(deep=False)
    if edge_mask is not None:
        da.masks['edges'] = edge_mask
    if holder_mask is not None:
        da.masks['holder_mask'] = holder_mask
    return MaskedData[ScatteringRunType](da)


providers = (detector_edge_mask, sample_holder_mask, mask_detectors)


def Sans2dWorkflow() -> sciline.Pipeline:
    """Create Sans2d workflow with default parameters."""
    from . import providers as isis_providers

    params = default_parameters()
    sans2d_providers = sans_providers + isis_providers + mantid_providers + providers
    return sciline.Pipeline(providers=sans2d_providers, params=params)


def Sans2dTutorialWorkflow() -> sciline.Pipeline:
    """
    Create Sans2d tutorial workflow.

    Equivalent to :func:`Sans2dWorkflow`, but with loaders for tutorial data instead
    of Mantid-based loaders.
    """
    workflow = Sans2dWorkflow()
    workflow.insert(load_tutorial_run)
    workflow.insert(load_tutorial_direct_beam)
    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import importlib.metadata

from . import data, general, io
from .general import LokiAtLarmorWorkflow, default_parameters

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

providers = general.providers + io.providers

del importlib

__all__ = [
    'data',
    'general',
    'io',
    'providers',
    'default_parameters',
    'LokiAtLarmorWorkflow',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
Providers for the ISIS instruments.
"""
import scipp as sc

from ..sans.types import (
    ConfiguredReducibleDataData,
    ConfiguredReducibleMonitor,
    CorrectForGravity,
    DetectorPixelShape,
    DimsToKeep,
    Incident,
    LabFrameTransform,
    MonitorType,
    NeXusMonitorName,
    RawData,
    RawMonitor,
    RunNumber,
    RunTitle,
    RunType,
    SampleRun,
    ScatteringRunType,
    TofData,
    TofMonitor,
    Transmission,
)
from .components import DetectorBankOffset, MonitorOffset, SampleOffset
from .data import LoadedFileContents


def default_parameters() -> dict:
    return {
        CorrectForGravity: False,
        DimsToKeep: tuple(),
        MonitorOffset[Incident]: MonitorOffset(sc.vector([0, 0, 0], unit='m')),
        MonitorOffset[Transmission]: MonitorOffset(sc.vector([0, 0, 0], unit='m')),
        DetectorBankOffset: DetectorBankOffset(sc.vector([0, 0, 0], unit='m')),
        SampleOffset: SampleOffset(sc.vector([0, 0, 0], unit='m')),
    }


def get_detector_data(
    dg: LoadedFileContents[RunType],
) -> RawData[RunType]:
    return RawData[RunType](dg['data'])


def get_monitor_data(
    dg: LoadedFileContents[RunType], nexus_name: NeXusMonitorName[MonitorType]
) -> RawMonitor[RunType, MonitorType]:
    # See https://github.com/scipp/sciline/issues/52 why copy needed
    mon = dg['monitors'][nexus_name]['data'].copy()
    return RawMonitor[RunType, MonitorType](mon)


def data_to_tof(
    da: ConfiguredReducibleDataData[ScatteringRunType],
) -> TofData[ScatteringRunType]:
    """Dummy conversion of data to time-of-flight data.
    The data already has a time-of-flight coordinate."""
    return TofData[ScatteringRunType](da)


def monitor_to_tof(
    da: ConfiguredReducibleMonitor[RunType, MonitorType]
) -> TofMonitor[RunType, MonitorType]:
    """Dummy conversion of monitor data to time-of-flight data.
    The monitor data already has a time-of-flight coordinate."""
    return TofMonitor[RunType, MonitorType](da)


def run_number(dg: LoadedFileContents[SampleRun]) -> RunNumber:
    """Get the run number from the raw sample data."""
    return RunNumber(int(dg['run_number']))


def run_title(dg: LoadedFileContents[SampleRun]) -> RunTitle:
    """Get the run title from the raw sample data."""
    return RunTitle(dg['run_title'].value)


def helium3_tube_detector_pixel_shape() -> DetectorPixelShape[ScatteringRunType]:
    # Pixel radius and length
    # found here:
    # https://github.com/mantidproject/mantid/blob/main/instrument/SANS2D_Definition_Tubes.xml
    R = 0.00405
    L = 0.002033984375
    pixel_shape = sc.DataGroup(
        {
            'vertices': sc.vectors(
                dims=['vertex'],
                values=[
                    # Coordinates in pixel-local coordinate system
                    # Bottom face center
                    [0, 0, 0],
                    # Bottom face edge
                    [R, 0, 0],
                    # Top face center
                    [0, L, 0],
                ],
                unit='m',
            ),
            'nexus_class': 'NXcylindrical_geometry',
        }
    )
    return pixel_shape


def lab_frame_transform() -> LabFrameTransform[ScatteringRunType]:
    # Rotate +y to -x
    return sc.spatial.rotation(value=[0, 0, 1 / 2**0.5, 1 / 2**0.5])


providers = (
    get_detector_data,
    get_monitor_data,
    data_to_tof,
    monitor_to_tof,
    run_number,
    run_title,
    lab_frame_transform,
    helium3_tube_detector_pixel_shape,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from typing import NewType

import sciline
import scipp as sc

from ..sans.types import (
    ConfiguredReducibleDataData,
    ConfiguredReducibleMonitor,
    MonitorType,
    RawData,
    RawMonitor,
    RunType,
    ScatteringRunType,
)


class MonitorOffset(sciline.Scope[MonitorType, sc.Variable], sc.Variable):
    """Offset for monitor position"""


SampleOffset = NewType('SampleOffset', sc.Variable)
DetectorBankOffset = NewType('DetectorBankOffset', sc.Variable)


def apply_component_user_offsets_to_raw_data(
    data: RawData[ScatteringRunType],
    sample_offset: SampleOffset,
    detector_bank_offset: DetectorBankOffset,
) -> ConfiguredReducibleDataData[ScatteringRunType]:
    """Apply user offsets to raw data.

    Parameters
    ----------
    data:
        Raw data.
    sample_offset:
        Sample offset.
    detector_bank_offset:
        Detector bank offset.
    """
    data = data.copy(deep=False)
    sample_pos = data.coords['sample_position']
    data.coords['sample_position'] = sample_pos + sample_offset.to(
        unit=sample_pos.unit, copy=False
    )
    pos = data.coords['position']
    data.coords['position'] = pos + detector_bank_offset.to(unit=pos.unit, copy=False)
    return ConfiguredReducibleDataData[ScatteringRunType](data)


def apply_component_user_offsets_to_raw_monitor(
    monitor_data: RawMonitor[RunType, MonitorType],
    monitor_offset: MonitorOffset[MonitorType],
) -> ConfiguredReducibleMonitor[RunType, MonitorType]:
    """Apply user offsets to raw monitor.
    Parameters
    ----------
    monitor_data:
        Raw monitor data.
    monitor_offset:
        Offset to apply to monitor position.
    """
    monitor_data = monitor_data.copy(deep=False)
    pos = monitor_data.coords['position']
    monitor_data.coords['position'] = pos + monitor_offset.to(unit=pos.unit, copy=False)
    return ConfiguredReducibleMonitor[RunType, MonitorType](monitor_data)


providers = (
    apply_component_user_offsets_to_raw_data,
    apply_component_user_offsets_to_raw_monitor,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pytest

from ess import isissans, loki, sans


@pytest.mark.parametrize('pkg', [sans, loki, isissans])
def test_has_version(pkg):
    assert hasattr(pkg, '__version__')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import uuid
from typing import Optional

import sciline
import scipp as sc
from scipp.scipy.interpolate import interp1d

from .common import mask_range
from .logging import get_logger
from .types import (
    BackgroundRun,
    BackgroundSubtractedIofQ,
    BackgroundSubtractedIofQxy,
    CleanDirectBeam,
    CleanMonitor,
    CleanQ,
    CleanQxy,
    CleanSummedQ,
    CleanSummedQMergedBanks,
    CleanSummedQxy,
    CleanSummedQxyMergedBanks,
    DimsToKeep,
    DirectBeam,
    Filename,
    FinalSummedQ,
    FinalSummedQxy,
    IofQ,
    IofQPart,
    IofQxy,
    MonitorType,
    NeXusDetectorName,
    NonBackgroundWavelengthRange,
    QBins,
    QxBins,
    QyBins,
    ReturnEvents,
    RunType,
    SampleRun,
    ScatteringRunType,
    UncertaintyBroadcastMode,
    WavelengthBins,
    WavelengthMonitor,
)
from .uncertainty import broadcast_with_upper_bound_variances


def preprocess_monitor_data(
    monitor: WavelengthMonitor[RunType, MonitorType],
    wavelength_bins: WavelengthBins,
    non_background_range: Optional[NonBackgroundWavelengthRange],
    uncertainties: UncertaintyBroadcastMode,
) -> CleanMonitor[RunType, MonitorType]:
    """
    Prepare monitor data for computing the transmission fraction.
    The input data are first converted to wavelength (if needed).
    If a ``non_background_range`` is provided, it defines the region where data is
    considered not to be background, and regions outside are background. A mean
    background level will be computed from the background and will be subtracted from
    the non-background counts.
    Finally, if wavelength bins are provided, the data is rebinned to match the
    requested binning.

    Parameters
    ----------
    monitor:
        The monitor to be pre-processed.
    wavelength_bins:
        The binning in wavelength to use for the rebinning.
    non_background_range:
        The range of wavelengths that defines the data which does not constitute
        background. Everything outside this range is treated as background counts.
    uncertainties:
        The mode for broadcasting uncertainties. See
        :py:class:`UncertaintyBroadcastMode` for details.

    Returns
    -------
    :
        The input monitors converted to wavelength, cleaned of background counts, and
        rebinned to the requested wavelength binning.
    """
    background = None
    if non_background_range is not None:
        mask = sc.DataArray(
            data=sc.array(dims=[non_background_range.dim], values=[True]),
            coords={non_background_range.dim: non_background_range},
        )
        background = mask_range(monitor, mask=mask).mean()

    if monitor.bins is not None:
        monitor = monitor.hist(wavelength=wavelength_bins)
    else:
        monitor = monitor.rebin(wavelength=wavelength_bins)

    if background is not None:
        if uncertainties == UncertaintyBroadcastMode.drop:
            monitor -= sc.values(background)
        elif uncertainties == UncertaintyBroadcastMode.upper_bound:
            monitor -= broadcast_with_upper_bound_variances(
                background, sizes=monitor.sizes
            )
        else:
            monitor -= background
    return CleanMonitor(monitor)


def resample_direct_beam(
    direct_beam: DirectBeam, wavelength_bins: WavelengthBins
) -> CleanDirectBeam:
    """
    If the wavelength binning of the direct beam function does not match the requested
    ``wavelength_bins``, perform a 1d interpolation of the function onto the bins.

    Parameters
    ----------
    direct_beam:
        The DataArray containing the direct beam function (it should have a dimension
        of wavelength).
    wavelength_bins:
        The binning in wavelength that the direct beam function should be resampled to.

    Returns
    -------
    :
        The direct beam function resampled to the requested resolution.
    """
    if sc.identical(direct_beam.coords['wavelength'], wavelength_bins):
        return direct_beam
    if direct_beam.variances is not None:
        logger = get_logger('sans')
        logger.warning(
            'An interpolation is being performed on the direct_beam function. '
            'The variances in the direct_beam function will be dropped.'
        )
    func = interp1d(
        sc.values(direct_beam),
        'wavelength',
        fill_value="extrapolate",
        bounds_error=False,
    )
    return CleanDirectBeam(func(wavelength_bins, midpoints=True))


def bin_in_q(
    data: CleanQ[ScatteringRunType, IofQPart],
    q_bins: QBins,
    dims_to_keep: DimsToKeep,
) -> CleanSummedQ[ScatteringRunType, IofQPart]:
    """
    Merges data from all pixels into a single I(Q) spectrum:

    * In the case of event data, events in all bins are concatenated
    * In the case of dense data, counts in all spectra are summed

    Parameters
    ----------
    data:
        A DataArray containing the data that is to be converted to Q.
    q_bins:
        The binning in Q to be used.
    dims_to_keep:
        Dimensions that should not be reduced and thus still be present in the final
        I(Q) result (this is typically the layer dimension).

    Returns
    -------
    :
        The input data converted to Q and then summed over all detector pixels.
    """
    out = _bin_in_q(data=data, edges={'Q': q_bins}, dims_to_keep=dims_to_keep)
    return CleanSummedQ[ScatteringRunType, IofQPart](out)


def bin_in_qxy(
    data: CleanQxy[ScatteringRunType, IofQPart],
    qx_bins: QxBins,
    qy_bins: QyBins,
    dims_to_keep: DimsToKeep,
) -> CleanSummedQxy[ScatteringRunType, IofQPart]:
    """
    Merges data from all pixels into a single I(Q) spectrum:

    * In the case of event data, events in all bins are concatenated
    * In the case of dense data, counts in all spectra are summed

    Parameters
    ----------
    data:
        A DataArray containing the data that is to be converted to Q.
    qx_bins:
        The binning in Qx to be used.
    qy_bins:
        The binning in Qy to be used.
    dims_to_keep:
        Dimensions that should not be reduced and thus still be present in the final
        I(Q) result (this is typically the layer dimension).

    Returns
    -------
    :
        The input data converted to Qx and Qy and then summed over all detector pixels.
    """
    # We make Qx the inner dim, such that plots naturally show Qx on the x-axis.
    out = _bin_in_q(
        data=data,
        edges={'Qy': qy_bins, 'Qx': qx_bins},
        dims_to_keep=dims_to_keep,
    )
    return CleanSummedQxy[ScatteringRunType, IofQPart](out)


def _bin_in_q(
    data: sc.DataArray, edges: dict[str, sc.Variable], dims_to_keep: tuple[str, ...]
) -> sc.DataArray:
    dims_to_reduce = set(data.dims) - {'wavelength'}
    if dims_to_keep is not None:
        dims_to_reduce -= set(dims_to_keep)

    if data.bins is not None:
        q_all_pixels = data.bins.concat(dims_to_reduce)
        # q_all_pixels may just have a single bin now, which currently yields
        # inferior performance when binning (no/bad multi-threading?).
        # We operate on the content buffer for better multi-threaded performance.
        if q_all_pixels.ndim == 0:
            content = q_all_pixels.bins.constituents['data']
            out = content.bin(**edges).assign_coords(q_all_pixels.coords)
        else:
            out = q_all_pixels.bin(**edges)
    else:
        # We want to flatten data to make histogramming cheaper (avoiding allocation of
        # large output before summing). We strip unnecessary content since it makes
        # flattening more expensive.
        stripped = data.copy(deep=False)
        for name, coord in data.coords.items():
            if (
                name not in {'Q', 'Qx', 'Qy', 'wavelength'}
                and set(coord.dims) & dims_to_reduce
            ):
                del stripped.coords[name]
        to_flatten = [dim for dim in data.dims if dim in dims_to_reduce]

        # Make dims to flatten contiguous, keep wavelength as the last dim
        data_dims = list(stripped.dims)
        for dim in to_flatten + ['wavelength']:
            data_dims.remove(dim)
            data_dims.append(dim)
        stripped = stripped.transpose(data_dims)
        # Flatten to helper dim such that `hist` will turn this into the new Q dim(s).
        # For sc.hist this has to be named 'Q'.
        helper_dim = 'Q'
        flat = stripped.flatten(dims=to_flatten, to=helper_dim)

        if len(edges) == 1:
            out = flat.hist(**edges)
        else:
            # sc.hist (or the underlying sc.bin) cannot deal with extra data dims,
            # work around by flattening and regrouping.
            for dim in flat.dims:
                if dim == helper_dim:
                    continue
                if dim not in flat.coords:
                    flat.coords[dim] = sc.arange(dim, flat.sizes[dim])
            out = (
                flat.flatten(to=str(uuid.uuid4()))
                .group(*[flat.coords[dim] for dim in flat.dims if dim != helper_dim])
                .drop_coords(dims_to_keep or ())
                .hist(**edges)
            )
    return out.squeeze()


def no_bank_merge(
    data: CleanSummedQ[ScatteringRunType, IofQPart]
) -> CleanSummedQMergedBanks[ScatteringRunType, IofQPart]:
    return CleanSummedQMergedBanks[ScatteringRunType, IofQPart](data)


def no_bank_merge_xy(
    data: CleanSummedQxy[ScatteringRunType, IofQPart]
) -> CleanSummedQxyMergedBanks[ScatteringRunType, IofQPart]:
    return CleanSummedQxyMergedBanks[ScatteringRunType, IofQPart](data)


def no_run_merge(
    data: CleanSummedQMergedBanks[ScatteringRunType, IofQPart]
) -> FinalSummedQ[ScatteringRunType, IofQPart]:
    return FinalSummedQ[ScatteringRunType, IofQPart](data)


def no_run_merge_xy(
    data: CleanSummedQxyMergedBanks[ScatteringRunType, IofQPart]
) -> FinalSummedQxy[ScatteringRunType, IofQPart]:
    return FinalSummedQxy[ScatteringRunType, IofQPart](data)


def _merge_contributions(data: list[sc.DataArray]) -> sc.DataArray:
    if len(data) == 1:
        return data[0]
    reducer = sc.reduce(data)
    return reducer.bins.concat() if data[0].bins is not None else reducer.sum()


def merge_banks(
    banks: sciline.Series[NeXusDetectorName, CleanSummedQ[ScatteringRunType, IofQPart]]
) -> CleanSummedQMergedBanks[ScatteringRunType, IofQPart]:
    """
    Merge the events or counts from multiple detector banks into a single numerator or
    denominator, before the normalization step.
    """
    return CleanSummedQMergedBanks[ScatteringRunType, IofQPart](
        _merge_contributions(list(banks.values()))
    )


def merge_banks_xy(
    banks: sciline.Series[
        NeXusDetectorName, CleanSummedQxy[ScatteringRunType, IofQPart]
    ]
) -> CleanSummedQxyMergedBanks[ScatteringRunType, IofQPart]:
    """
    Merge the events or counts from multiple detector banks into a single numerator or
    denominator, before the normalization step.
    """
    return CleanSummedQxyMergedBanks[ScatteringRunType, IofQPart](
        _merge_contributions(list(banks.values()))
    )


def merge_runs(
    runs: sciline.Series[
        Filename[ScatteringRunType],
        CleanSummedQMergedBanks[ScatteringRunType, IofQPart],
    ],
) -> FinalSummedQ[ScatteringRunType, IofQPart]:
    """
    Merge the events or counts from multiple runs into a single numerator or
    denominator, before the normalization step.
    """
    return FinalSummedQ[ScatteringRunType, IofQPart](
        _merge_contributions(list(runs.values()))
    )


def merge_runs_xy(
    runs: sciline.Series[
        Filename[ScatteringRunType],
        CleanSummedQxyMergedBanks[ScatteringRunType, IofQPart],
    ],
) -> FinalSummedQxy[ScatteringRunType, IofQPart]:
    """
    Merge the events or counts from multiple runs into a single numerator or
    denominator, before the normalization step.
    """
    return FinalSummedQxy[ScatteringRunType, IofQPart](
        _merge_contributions(list(runs.values()))
    )


def _subtract_background(
    sample: sc.DataArray,
    background: sc.DataArray,
    return_events: ReturnEvents,
) -> sc.DataArray:
    if return_events and sample.bins is not None and background.bins is not None:
        return sample.bins.concatenate(-background)
    if sample.bins is not None:
        sample = sample.bins.sum()
    if background.bins is not None:
        background = background.bins.sum()
    return sample - background


def subtract_background(
    sample: IofQ[SampleRun],
    background: IofQ[BackgroundRun],
    return_events: ReturnEvents,
) -> BackgroundSubtractedIofQ:
    return BackgroundSubtractedIofQ(
        _subtract_background(
            sample=sample, background=background, return_events=return_events
        )
    )


def subtract_background_xy(
    sample: IofQxy[SampleRun],
    background: IofQxy[BackgroundRun],
    return_events: ReturnEvents,
) -> BackgroundSubtractedIofQxy:
    return BackgroundSubtractedIofQxy(
        _subtract_background(
            sample=sample, background=background, return_events=return_events
        )
    )


providers = (
    preprocess_monitor_data,
    resample_direct_beam,
    bin_in_q,
    bin_in_qxy,
    subtract_background,
    subtract_background_xy,
    no_bank_merge,
    no_bank_merge_xy,
    no_run_merge,
    no_run_merge_xy,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Callable, List

import scipp as sc

from ess import loki, sans
from ess.sans.types import (
    BackgroundRun,
    CorrectForGravity,
    EmptyBeamRun,
    Filename,
    NeXusDetectorName,
    QBins,
    QxBins,
    QyBins,
    ReturnEvents,
    SampleRun,
    TransmissionRun,
    UncertaintyBroadcastMode,
    WavelengthBins,
)


def make_params() -> dict:
    params = loki.default_parameters()

    params[NeXusDetectorName] = 'larmor_detector'
    params[Filename[SampleRun]] = loki.data.loki_tutorial_sample_run_60339()
    params[Filename[BackgroundRun]] = loki.data.loki_tutorial_background_run_60393()
    params[
        Filename[TransmissionRun[SampleRun]]
    ] = loki.data.loki_tutorial_sample_transmission_run()
    params[
        Filename[TransmissionRun[BackgroundRun]]
    ] = loki.data.loki_tutorial_run_60392()
    params[Filename[EmptyBeamRun]] = loki.data.loki_tutorial_run_60392()

    params[WavelengthBins] = sc.linspace(
        'wavelength', start=1.0, stop=13.0, num=51, unit='angstrom'
    )
    params[CorrectForGravity] = True
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    params[ReturnEvents] = False

    params[QxBins] = sc.linspace('Qx', start=-0.3, stop=0.3, num=91, unit='1/angstrom')
    params[QyBins] = sc.linspace('Qy', start=-0.2, stop=0.3, num=78, unit='1/angstrom')
    params[QBins] = sc.linspace('Q', start=0.01, stop=0.3, num=101, unit='1/angstrom')

    return params


def loki_providers_no_beam_center_finder() -> List[Callable]:
    from ess.isissans.io import read_xml_detector_masking

    return list(
        sans.providers
        + loki.providers
        + (
            read_xml_detector_masking,
            loki.io.dummy_load_sample,
        )
    )


def loki_providers() -> List[Callable]:
    return loki_providers_no_beam_center_finder() + [
        sans.beam_center_finder.beam_center_from_center_of_mass
    ]
